#ifndef Bp2Fits_h
#define Bp2Fits_h

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of src
//
// src is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// src is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <map>
#include <pwd.h>
#include <string>
#include <vector>
#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>
#include <funtools.h>
#include <suplibxx/colselect.h>
#include <suplibxx/str.h>

#include "Bp2FitsErr.h"

typedef enum
{
  DT_int,
  DT_uint,
  DT_double,
  DT_char
} DataType;


typedef struct
{
  std::string  mode;
  int    offset;

  // this is where the data originate in the bpipe core image
  void*  src;

  // this is where Funtools will find the data. it may be the
  // same as src, or may point at allocated space if the data are in a
  // matrix
  void*  data;

  // the minimum and maximum values in the column.  these point at
  // allocated space.
  void*  min;
  void*  max;

  // C++ datatype
  DataType cxxdt;

  bool   allocate;

  // FITS data type
  char   dtype;

  std::string scale;
  std::string  zero;

  std::string  type;
  std::string  form;
  std::string  unit;
  std::string  null;
  std::string  disp;
  std::string  dim;
  std::string  lmin;
  std::string  lmax;
  std::string  dmin;
  std::string  dmax;

  size_t  nelem;
  size_t  stride;
} ColumnInfo;

typedef struct
{
  std::string fitsfield;
  std::string unit;
  std::string null;
  std::string scale;
  std::string zero;
  std::string disp;
  std::string lmin;
  std::string lmax;
  std::string dmin;
  std::string dmax;
} OverrideInfo;

struct Comment {

  public:
    std::string keyword;
    std::string value;
    std::string text;

  Comment( const std::string& keyword_, const std::string& value_, const std::string& text_ )
      : keyword(keyword_), value(value_), text(text_) {

      if ( 0 == keyword.size( ) ) {
	  keyword = "COMMENT";
	  value   = text;
      }

  }
};

class Bp2Fits {
  
  template<typename Type>
    void convertDpktf2ColTYPE( 
      const char* bpname,
      const char fitsform,
      const char fitsdisp,
      DataType cxxdt,
      DpktField* dpktf,
      Type* bpdata,
      const size_t vecsize,
      std::string sfx[]
      );

  template<class Type>
    friend void transferDataTYPE( Bp2Fits& bp2fits, BPMatrix* bpmatrix, Type* bpdata, size_t vecsize, size_t elemsize );
  
 public:
  Bp2Fits( const std::string& fits, const std::string& bpipe, bool debug=false );
  ~Bp2Fits( void );

  void writeTable( void );
  void rewriteHDU( void );
  
  void trans(      const std::string& trans     ) { _trans        = trans;        };
  void debug(      bool debug      ) { _debug      = debug;      };
  void doublepass( bool doublepass ) { _doublepass = doublepass; };
  void longstrn(   bool longstrn   ) { _longstrn   = longstrn;   };
  void reserved(   bool reserved   ) { _reserved   = reserved;   };
  void vectors(    bool vectors    ) { _vectors    = vectors;    };
  void verbose(    bool verbose    ) { _verbose    = verbose;    };

  void add_exact_dpktfs( const std::string& dpktfs  ) { suplib::tok( _add_exact_dpktfs, dpktfs, ", "); };
  void add_regex_dpktfs( const std::string& dpktfs  ) { suplib::tok( _add_regex_dpktfs, dpktfs, ", "); };
  void del_exact_dpktfs( const std::string& dpktfs  ) { suplib::tok( _del_exact_dpktfs, dpktfs, ", "); };
  void del_regex_dpktfs( const std::string& dpktfs  ) { suplib::tok( _del_regex_dpktfs, dpktfs, ", "); };
  
 private:
  void readOverRide( void );
  void addColumns( void );
  void readwriteData( void );
  void writeHDU_longstrn( bool primary=false );
  void writeHDU_colinfo( void );
  void writeHDU_hdrf( void );
  void writeHDU_comments( bool primary=false );

  void createTable( void );
  void convertDpktf2Col(const char* bpname, DpktField* dpktf);

  void bpcheck( void );
  
  BPipe* _bpipe;
  void* _dpkt;
  bool _empty;
  
  Fun _fun;
  std::vector<ColumnInfo*> _ci;
  std::map<std::string,OverrideInfo*> _orinfo;

  std::vector<std::string> _ifields;
  std::vector<std::string> _ofields;

  std::vector<Comment> _comments;
  
  std::string _trans;

  void** _data;
  size_t _ndpktfs;
  size_t _ncols;
  size_t _nrows;

  size_t _precision;
  
  bool _debug;
  bool _doublepass;
  bool _longstrn;
  bool _reserved;
  bool _vectors;
  bool _verbose;

  std::vector<std::string> _add_exact_dpktfs;
  std::vector<std::string> _add_regex_dpktfs;
  std::vector<std::string> _del_exact_dpktfs;
  std::vector<std::string> _del_regex_dpktfs;
};

#endif
